#!/bin/sh

help_screen() {
  echo "Usage: $0 [-s FILE] [-w DIR]"
  echo "Turns a POSIX shell script into a Windows Executable."
  printf "\n"
  echo "Options:"
  echo "-s=FILE         Script to convert to executable."
  echo "-w=DIR          Directory to use as a work directory."
  echo "-n              Do not open console when executable is ran. Useful for non-interactive scripts."
  echo "-h              Shows this screen."
  echo "-c              Cleans workspace."
  echo "-l              Shows license of this script."
  printf "\n"
  echo "Example:"
  echo "# Converts helloworld.sh to an executable while using a different workspace by default."
  echo "$0 -s helloworld.sh -w /tmp/sh2exe_work"
  printf "\n"
  echo "Exit status:"
  echo "0 = Success."
  echo "1 = Initialization failure. (Before runtime was being built.)"
  echo "5 = Failed to change directory while runtime was building."
  printf "\n"
  echo "Website : https://sh2exe.iotib.net"
  echo "Git     : https://pagure.io/dalazyslackers/sh2exe"
  return 0
}

show_license() {
  echo "
  Copyright 2023 Charadon

  Licensed under the Apache License, Version 2.0 (the \"License\");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an \"AS IS\" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
  "
  return 0
}

clean() {
  set -u
  set +e
  rm -rvf "$WORKDIR"
  rm -fv "7zr.exe"
  rm -fv output.log
  make -C bin2c clean
  return 0
}


GUI=""
if [ -z "$1" ];
then
  help_screen
  exit 1
fi
CLEAN="false"
while getopts ":s:w:chln" opt;
do
  case "${opt}" in
    s)
      SCRIPT="${OPTARG}"
      if [ ! -f "$SCRIPT" ];
      then
        echo "File doesn't exist."
        exit 1
      fi
      SHEBANG="$(head -1 "$SCRIPT")"
      if [ ! "$SHEBANG" = "#!/bin/sh" ] && [ ! "$SHEBANG" = "#!/bin/bash" ];
      then
        echo "File has invalid shebang. Aborting."
        exit 1
      fi
      ;;
    w)
      WORKDIR="${OPTARG}"
      ;;
    h)
      help_screen
      exit 0
      ;;
    c)
      CLEAN="true"
      exit 0
      ;;
    l)
      show_license
      exit 0
      ;;
    n)
      GUI="-mwindows"
      ;;
    *)
      help_screen
      exit 1
      ;;
  esac
done

if [ "$CLEAN" = "true" ];
then
  clean
fi

if [ -z "$SCRIPT" ];
then
  help_screen
  exit 1
fi

set -eu

if [ ! "$MSYSTEM" = "MSYS" ];
then
  echo "You must run MSYS2 with the MSYS runtime."
	exit 1
fi

wait_for_finish() {
  printf "|"
  while kill -0 "$1" > /dev/null 2>&1;
  do
    printf "\b|"
    sleep 0.1
    printf "\b/"
    sleep 0.1
    printf "\b-"
    sleep 0.1
    printf "\b\\"
    sleep 0.1
  done
  printf "\b"
  wait "$1"
  return "$?"
}

# Create work directory.
WORKDIR="${WORKDIR:-$PWD/work}"
rm -rf "$WORKDIR"
mkdir -p "$WORKDIR/var/lib/pacman"

OUTPUT_LOG="${OUTPUT_LOG:-output.log}"
rm -f "$OUTPUT_LOG"

# Update system
printf "Updating system... "
yes | pacman -Syu >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
printf "\rUpdating system... ✓\n"

# Install dependencies
printf "Installing dependencies... "
MISSING_DEPS=""
{
  if ! pacman -Qi curl > /dev/null 2>&1;
  then
    MISSING_DEPS="$MISSING_DEPS curl"
  fi
  if ! pacman -Qi make > /dev/null 2>&1;
  then
    MISSING_DEPS="$MISSING_DEPS make"
  fi
  if ! pacman -Qi gcc > /dev/null 2>&1;
  then
    MISSING_DEPS="$MISSING_DEPS gcc"
  fi
  if ! pacman -Qi mingw-w64-x86_64-gcc > /dev/null 2>&1;
  then
    MISSING_DEPS="$MISSING_DEPS mingw-w64-x86_64-gcc"
  fi
  if [ ! $MISSING_DEPS = "" ];
  then
    yes | pacman -S $MISSING_DEPS
  fi
} >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
printf "\rInstalling dependencies... ✓\n"

# Download 7-zip
printf "Downloading 7-zip... "
curl "https://7-zip.org/a/7zr.exe" -O >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
printf "\rDownloading 7-zip... ✓\n"

# Compile bin2c to embed runtime and 7zip into source file.
printf "Building bin2c... "
cd bin2c || exit 5
make clean >> "$OUTPUT_LOG" 2>&1
make -j "$(nproc)" >> "$OUTPUT_LOG" 2>&1
cd .. || exit 5
printf "\rBuilding bin2c... ✓\n"

# Create runtime environment.
printf "Creating runtime environment... "
yes | pacman --root "$WORKDIR" -Sy >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
yes | pacman --root "$WORKDIR" -S \
  coreutils \
  msys2-runtime \
  grep \
  gawk \
  less \
  sed \
  mksh \
  nano \
  gzip \
  xz \
  bzip2 \
  tar \
  cpio \
  zip \
  unzip >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
cp "$SCRIPT" "$WORKDIR/usr/bin/"
printf "\rCreating runtime environment... ✓\n"

# Remove useless files from runtime environment.
printf "Pruning useless stuff from runtime... "
{
  rm -rf "${WORKDIR:?}/var"
  rm -rf "${WORKDIR:?}/usr/share/info"
  rm -rf "${WORKDIR:?}/usr/share/man"
  rm -rf "${WORKDIR:?}/usr/share/doc"
  rm -rf "${WORKDIR:?}/usr/share/licenses"
  rm -rf "${WORKDIR:?}/usr/share/locale"
  rm -rf "${WORKDIR:?}/usr/include"
  for i in $WORKDIR/usr/bin/*.exe;
  do
    strip --strip-unneeded "$i"
  done
  for i in $WORKDIR/usr/bin/*.dll;
  do
    strip --strip-unneeded "$i"
  done
} >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
printf "\rPruning useless stuff from runtime... ✓\n"
# Compress the runtime down with 7zip
printf "Create runtime archive... "
{
  # If shebang is the POSIX Shell (/bin/sh), use MirBSD Kornshell instead.
  if [ "$SHEBANG" = "#!/bin/sh" ];
  then
    rm "$WORKDIR/usr/bin/sh" # Remove bash
    cp "$WORKDIR/usr/bin/mksh" "$WORKDIR/usr/bin/sh" # Put mksh in it's place.
  fi
  ./7zr.exe -mx1 a "$WORKDIR"/runtime.7z "$WORKDIR/*"
} >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
printf "\rCreate runtime archive... ✓\n"

printf "Compiling executable... "
cp runsh.c "$WORKDIR/$(basename $SCRIPT).c"
sed -i "s/INSERT_SCRIPT_NAME_HERE/$(basename $SCRIPT)/g" "$WORKDIR/$(basename $SCRIPT).c"
bin2c/bin2c.exe "7zr.exe" "$WORKDIR/7zip.h" SevenZip >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
bin2c/bin2c.exe "$WORKDIR/runtime.7z" "$WORKDIR/runtime.h" RuntimeArchive >> "$OUTPUT_LOG" 2>&1 &
wait_for_finish "$!"
MSYSTEM="MINGW64" bash -l -c "gcc -I$WORKDIR -Wall -O1 \"$WORKDIR/$(basename $SCRIPT).c\" $GUI -o \"$WORKDIR/$(basename $SCRIPT).exe\"" &
wait_for_finish "$!"
printf "\rCompiling executable... ✓\n"
echo "If everything went well, this script should've made an executable at:"
echo "$WORKDIR/$(basename $SCRIPT).exe"
